'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function withTypeSymbol(value) {
    return value;
}

const any = () => withTypeSymbol({ type: 'any' });
const isAnyType = (value) => value.type === 'any';
const validate = (_schema, _value, _path) => [];
const code = (_schema) => 'any';

const bigint = () => withTypeSymbol({ type: 'bigint' });
const isBigintType = (value) => value.type === 'bigint';
const validate$1 = (_schema, value, path) => typeof value === 'bigint' ? [] : [invalidTypeIssue('bigint', value, path)];
const code$1 = (_schema) => 'bigint';

const boolean = () => withTypeSymbol({ type: 'boolean' });
const isBooleanType = (value) => value.type === 'boolean';
const validate$2 = (_schema, value, path) => typeof value === 'boolean' ? [] : [invalidTypeIssue('boolean', value, path)];
const code$2 = (_schema) => 'boolean';

const functionType = (...args) => withTypeSymbol({ type: 'function', args });
const isFunctionType = (value) => value.type === 'function';
const validate$3 = (_schema, value, path) => (typeof value === 'function' ? [] : [invalidTypeIssue('function', value, path)]);
const code$3 = (schema) => {
    if (schema.args.length === 0) {
        throw new Error('Not enough arguments');
    }
    if (schema.args.length === 1) {
        return `() => ${_code(schema.args[0])}`;
    }
    const argTypes = schema.args.slice(0, schema.args.length - 1);
    const args = argTypes.map((arg, idx) => `arg_${idx}: ${_code(arg)}`);
    const returnType = schema.args[schema.args.length - 1];
    return `(${args.join(', ')}) => ${_code(returnType)}`;
};

const isModifiedType = (value) => value.type === 'modified';

const readonly = (item) => {
    if (isReadonlyType(item)) {
        return item;
    }
    if (isOptionalType(item)) {
        const wrapped = { ...item, readonly: true };
        return wrapped;
    }
    return withTypeSymbol({ type: 'modified', readonly: true, item });
};
const isReadonlyType = (value) => isModifiedType(value) && value.readonly === true;
const validate$4 = (schema, value, path) => _validate(schema.item, value, path);
const code$4 = (schema) => _code(schema.item);

const optional = (item) => {
    if (isOptionalType(item)) {
        return item;
    }
    if (isReadonlyType(item)) {
        const wrapped = { ...item, optional: true };
        return wrapped;
    }
    return withTypeSymbol({ type: 'modified', optional: true, item });
};
const isOptionalType = (value) => isModifiedType(value) && value.optional === true;
const validate$5 = (schema, value, path) => (value === undefined ? [] : _validate(schema.item, value, path));
const code$5 = (schema) => `(${_code(schema.item)} | undefined)`;

const object = (properties, strict = true) => withTypeSymbol({ type: 'object', properties, strict });
const isObjectType = (value) => value.type === 'object';
const validate$6 = (schema, value, path) => {
    if (typeof value !== 'object' || value === null) {
        return [invalidTypeIssue('object', value, path)];
    }
    const schemaKeys = new Set(Object.keys(schema.properties));
    const valueKeys = new Set(Object.keys(value));
    const extraKeys = new Set([...valueKeys].filter((key) => !schemaKeys.has(key)));
    const requiredKeys = new Set([...schemaKeys].filter((key) => !isOptionalType(schema.properties[key])));
    const missingKeys = new Set([...requiredKeys].filter((key) => !valueKeys.has(key)));
    const issues = [];
    if (schema.strict && extraKeys.size > 0) {
        issues.push({
            type: 'INVALID_VALUE',
            message: `Value provided unexpected keys: ${[...extraKeys].join(', ')}`,
            path: pathToString(path),
        });
    }
    if (missingKeys.size > 0) {
        issues.push({
            type: 'INVALID_VALUE',
            message: `Value is missing required keys: ${[...missingKeys].join(', ')}`,
            path: pathToString(path),
        });
    }
    issues.push(...[...valueKeys].flatMap((key) => schemaKeys.has(key)
        ? _validate(schema.properties[key], value[key], [...path, key])
        : []));
    return issues;
};
const code$6 = (schema) => {
    const props = [];
    for (const prop of Object.keys(schema.properties)) {
        const propSchema = schema.properties[prop];
        const isOptional = isOptionalType(propSchema);
        const isReadonly = isReadonlyType(propSchema);
        props.push(`${isReadonly ? 'readonly ' : ''}"${prop}"${isOptional ? '?' : ''}: ${_code(propSchema)}`);
    }
    return `{${props.join('; ')}}`;
};

const intersection = (left, right) => withTypeSymbol({ type: 'intersection', left, right });
const isIntersectionType = (value) => value.type === 'intersection';
function mergeIntersectionTypes(left, right) {
    if (!isObjectType(left) || !isObjectType(right)) {
        return [left, right];
    }
    const strict = left.strict && right.strict;
    const mergedObjectType = withTypeSymbol({ type: 'object', properties: {}, strict });
    for (const objectType of [left, right]) {
        const keys = Object.keys(objectType.properties);
        for (const key of keys) {
            if (mergedObjectType.properties[key] === undefined) {
                mergedObjectType.properties[key] = objectType.properties[key];
            }
            else {
                mergedObjectType.properties[key] = intersection(mergedObjectType.properties[key], objectType.properties[key]);
            }
        }
    }
    return [mergedObjectType];
}
const validate$7 = (schema, value, path) => mergeIntersectionTypes(schema.left, schema.right).flatMap((candidate) => _validate(candidate, value, path));
const code$7 = (schema) => `(${_code(schema.left)} & ${schema.right})`;

const literal = (value) => withTypeSymbol({ type: 'literal', value });
const isLiteralType = (value) => value.type === 'literal';
const equalsLiteral = (literal, value, path) => {
    if (Array.isArray(literal)) {
        if (!Array.isArray(value)) {
            return [invalidTypeIssue(literal.toString(), value, path)];
        }
        const issues = [];
        if (literal.length !== value.length) {
            issues.push({
                type: 'INVALID_VALUE',
                message: `Expected tuple of length ${literal.length}, got one of length ${value.length}`,
                path: pathToString(path),
            });
        }
        issues.push(...value.flatMap((item, idx) => idx < literal.length ? equalsLiteral(literal[idx], item, [...path, idx.toString()]) : []));
        return issues;
    }
    if (literal === null) {
        return value === null ? [] : [invalidTypeIssue('null', value, path)];
    }
    if (literal === undefined) {
        return value === undefined ? [] : [invalidTypeIssue('undefined', value, path)];
    }
    if (typeof literal === 'object') {
        if (typeof value !== 'object' || value === null) {
            return [invalidTypeIssue(literal.toString(), value, path)];
        }
        const literalKeys = new Set(Object.keys(literal));
        const valueKeys = new Set(Object.keys(value));
        const extraKeys = new Set([...valueKeys].filter((key) => !literalKeys.has(key)));
        const missingKeys = new Set([...literalKeys].filter((key) => !valueKeys.has(key)));
        const issues = [];
        if (extraKeys.size > 0) {
            issues.push({
                type: 'INVALID_VALUE',
                message: `Value provided unexpected keys: ${[...extraKeys].join(', ')}`,
                path: pathToString(path),
            });
        }
        if (missingKeys.size > 0) {
            issues.push({
                type: 'INVALID_VALUE',
                message: `Value is missing required keys: ${[...missingKeys].join(', ')}`,
                path: pathToString(path),
            });
        }
        issues.push(...[...valueKeys].flatMap((key) => literalKeys.has(key)
            ? equalsLiteral(literal[key], value[key], [...path, key])
            : []));
        return issues;
    }
    return literal === value ? [] : [invalidTypeIssue(literal.toString(), value, path)];
};
const validate$8 = (schema, value, path) => equalsLiteral(schema.value, value, path);
const code$8 = (schema) => JSON.stringify(schema.value);

const literalUnion = (...values) => withTypeSymbol({ type: 'literalUnion', values });
const isLiteralUnionType = (value) => value.type === 'literalUnion';
const validate$9 = (schema, value, path) => {
    const candidateIssues = [];
    for (const candidate of schema.values) {
        const issues = equalsLiteral(candidate, value, path);
        if (issues.length === 0) {
            return [];
        }
        candidateIssues.push(issues);
    }
    return [
        {
            type: 'NO_UNION_TYPE_MATCH',
            message: `Value did not match any union type candidates`,
            path: pathToString(path),
            candidateIssues,
        },
    ];
};
const code$9 = (schema) => `(${schema.values.map((value) => JSON.stringify(value)).join(' | ')})`;

const nullType = () => withTypeSymbol({ type: 'null' });
const isNullType = (value) => value.type === 'null';
const validate$a = (_schema, value, path) => value === null ? [] : [invalidTypeIssue('null', value, path)];
const code$a = (_schema) => 'null';

const number = (options = {}) => withTypeSymbol({ type: 'number', options });
const isNumberType = (value) => value.type === 'number';
const validate$b = (schema, value, path) => {
    if (typeof value !== 'number') {
        return [invalidTypeIssue('number', value, path)];
    }
    const issues = [];
    if (schema.options.multipleOf !== undefined && value % schema.options.multipleOf !== 0) {
        issues.push({
            type: 'INVALID_VALUE',
            message: `Value must be a multiple of ${schema.options.multipleOf}`,
            path: pathToString(path),
        });
    }
    if (schema.options.maximum !== undefined && value > schema.options.maximum) {
        issues.push({
            type: 'INVALID_VALUE',
            message: `Value must be less than or equal to ${schema.options.maximum}, instead was ${value}`,
            path: pathToString(path),
        });
    }
    if (schema.options.exclusiveMaximum !== undefined && value >= schema.options.exclusiveMaximum) {
        issues.push({
            type: 'INVALID_VALUE',
            message: `Value must be strictly less than ${schema.options.exclusiveMaximum}, instead was ${value}`,
            path: pathToString(path),
        });
    }
    if (schema.options.minimum !== undefined && value < schema.options.minimum) {
        issues.push({
            type: 'INVALID_VALUE',
            message: `Value must be greater than or equal to ${schema.options.minimum}, instead was ${value}`,
            path: pathToString(path),
        });
    }
    if (schema.options.exclusiveMinimum !== undefined && value <= schema.options.exclusiveMinimum) {
        issues.push({
            type: 'INVALID_VALUE',
            message: `Value must be strictly greater than ${schema.options.exclusiveMinimum}, instead was ${value}`,
            path: pathToString(path),
        });
    }
    return issues;
};
const code$b = (_schema) => 'number';

const record = (type) => withTypeSymbol({ type: 'record', additionalProperties: type });
const isRecordType = (value) => value.type === 'record';
const validate$c = (schema, value, path) => {
    if (typeof value !== 'object' || value === null) {
        return [invalidTypeIssue('object', value, path)];
    }
    return Object.keys(value).flatMap((key) => _validate(schema.additionalProperties, value[key], [...path, key]));
};
const code$c = (schema) => `Record<string, ${_code(schema.additionalProperties)}>`;

const string = (options = {}) => withTypeSymbol({ type: 'string', options });
const isStringType = (value) => value.type === 'string';
const validate$d = (schema, value, path) => {
    if (typeof value !== 'string') {
        return [invalidTypeIssue('string', value, path)];
    }
    const issues = [];
    const length = value.length;
    if (schema.options.maxLength !== undefined && length > schema.options.maxLength) {
        issues.push({
            type: 'INVALID_VALUE',
            message: `String value must have a length no greater than ${schema.options.maxLength}, instead was ${length}`,
            path: pathToString(path),
        });
    }
    if (schema.options.minLength !== undefined && length < schema.options.minLength) {
        issues.push({
            type: 'INVALID_VALUE',
            message: `String value must have a length no less than ${schema.options.minLength}, instead was ${length}`,
            path: pathToString(path),
        });
    }
    if (schema.options.pattern !== undefined) {
        const regexp = new RegExp(schema.options.pattern);
        if (!regexp.test(value)) {
            issues.push({
                type: 'INVALID_VALUE',
                message: `String value ${JSON.stringify(value)} must match pattern: ${schema.options.pattern}`,
                path: pathToString(path),
            });
        }
    }
    return issues;
};
const code$d = (_schema) => 'string';

const tuple = (...items) => withTypeSymbol({ type: 'tuple', items });
const isTupleType = (value) => value.type === 'tuple';
const validate$e = (schema, value, path) => {
    if (!Array.isArray(value)) {
        return [invalidTypeIssue(JSON.stringify(schema.items), value, path)];
    }
    const issues = [];
    if (schema.items.length !== value.length) {
        issues.push({
            type: 'INVALID_VALUE',
            message: `Expected tuple of length ${schema.items.length}, got one of length ${value.length}`,
            path: pathToString(path),
        });
    }
    issues.push(...value.flatMap((item, idx) => idx < schema.items.length ? _validate(schema.items[idx], item, [...path, idx.toString()]) : []));
    return issues;
};
const code$e = (schema) => `[${schema.items.map((item) => _code(item)).join(', ')}]`;

const undefined$1 = () => withTypeSymbol({ type: 'undefined' });
const isUndefinedType = (value) => value.type === 'undefined';
const validate$f = (_schema, value, path) => typeof value === 'undefined' ? [] : [invalidTypeIssue('undefined', value, path)];
const code$f = (_schema) => 'undefined';

const union = (...items) => withTypeSymbol({ type: 'union', oneOf: items });
const isUnionType = (value) => value.type === 'union';
const validate$g = (schema, value, path) => {
    const candidateIssues = [];
    for (const candidate of schema.oneOf) {
        const issues = _validate(candidate, value, path);
        if (issues.length === 0) {
            return [];
        }
        candidateIssues.push(issues);
    }
    return [
        {
            type: 'NO_UNION_TYPE_MATCH',
            message: `Value did not match any union type candidates`,
            path: pathToString(path),
            candidateIssues,
        },
    ];
};
const code$g = (schema) => `(${schema.oneOf.map((item) => _code(item)).join(' | ')})`;

const unknown = () => withTypeSymbol({ type: 'unknown' });
const isUnknownType = (value) => value.type === 'unknown';
const validate$h = (_schema, _value, _path) => [];
const code$h = (_schema) => 'unknown';

const voidType = () => withTypeSymbol({ type: 'void' });
const isVoidType = (value) => value.type === 'void';
const validate$i = (_schema, value, path) => [
    { type: 'INVALID_VALUE', message: `Expected void, instead got ${value}`, path: pathToString(path) },
];
const code$i = (_schema) => 'void';

const _code = (schema) => {
    if (isAnyType(schema)) {
        return code();
    }
    if (isArrayType(schema)) {
        return code$k(schema);
    }
    if (isBigintType(schema)) {
        return code$1();
    }
    if (isBooleanType(schema)) {
        return code$2();
    }
    if (isFunctionType(schema)) {
        return code$3(schema);
    }
    if (isIntersectionType(schema)) {
        return code$7(schema);
    }
    if (isLiteralType(schema)) {
        return code$8(schema);
    }
    if (isLiteralUnionType(schema)) {
        return code$9(schema);
    }
    if (isNullType(schema)) {
        return code$a();
    }
    if (isNumberType(schema)) {
        return code$b();
    }
    if (isObjectType(schema)) {
        return code$6(schema);
    }
    if (isOptionalType(schema)) {
        return code$5(schema);
    }
    if (isReadonlyType(schema)) {
        return code$4(schema);
    }
    if (isRecordType(schema)) {
        return code$c(schema);
    }
    if (isStringType(schema)) {
        return code$d();
    }
    if (isTupleType(schema)) {
        return code$e(schema);
    }
    if (isUndefinedType(schema)) {
        return code$f();
    }
    if (isUnionType(schema)) {
        return code$g(schema);
    }
    if (isUnknownType(schema)) {
        return code$h();
    }
    if (isVoidType(schema)) {
        return code$i();
    }
    throw new Error('Invalid schema type');
};
const code$j = (schema) => _code(schema);

const array = (items) => withTypeSymbol({ type: 'array', items });
const isArrayType = (value) => value.type === 'array';
const validate$j = (schema, value, path) => {
    if (!Array.isArray(value)) {
        return [invalidTypeIssue('array', value, path)];
    }
    return value.flatMap((item, idx) => _validate(schema.items, item, [...path, idx.toString()]));
};
const code$k = (schema) => `Array<${_code(schema.items)}>`;

const pathToString = (path) => `/${path.join('/')}`;
const invalidTypeIssue = (expected, value, path) => ({
    type: 'INVALID_TYPE',
    message: `Invalid type, expected ${expected}, got ${typeof value} ${value}`,
    path: pathToString(path),
});
const _validate = (schema, value, path) => {
    if (isAnyType(schema)) {
        return validate();
    }
    if (isArrayType(schema)) {
        return validate$j(schema, value, path);
    }
    if (isBigintType(schema)) {
        return validate$1(schema, value, path);
    }
    if (isBooleanType(schema)) {
        return validate$2(schema, value, path);
    }
    if (isFunctionType(schema)) {
        return validate$3(schema, value, path);
    }
    if (isIntersectionType(schema)) {
        return validate$7(schema, value, path);
    }
    if (isLiteralType(schema)) {
        return validate$8(schema, value, path);
    }
    if (isLiteralUnionType(schema)) {
        return validate$9(schema, value, path);
    }
    if (isNullType(schema)) {
        return validate$a(schema, value, path);
    }
    if (isNumberType(schema)) {
        return validate$b(schema, value, path);
    }
    if (isObjectType(schema)) {
        return validate$6(schema, value, path);
    }
    if (isOptionalType(schema)) {
        return validate$5(schema, value, path);
    }
    if (isReadonlyType(schema)) {
        return validate$4(schema, value, path);
    }
    if (isRecordType(schema)) {
        return validate$c(schema, value, path);
    }
    if (isStringType(schema)) {
        return validate$d(schema, value, path);
    }
    if (isTupleType(schema)) {
        return validate$e(schema, value, path);
    }
    if (isUndefinedType(schema)) {
        return validate$f(schema, value, path);
    }
    if (isUnionType(schema)) {
        return validate$g(schema, value, path);
    }
    if (isUnknownType(schema)) {
        return validate$h();
    }
    if (isVoidType(schema)) {
        return validate$i(schema, value, path);
    }
    return [{ type: 'INVALID_SCHEMA', message: 'Invalid schema type', path: '' }];
};
const validate$k = (schema, value) => _validate(schema, value, []);

class SchemaTypeError extends Error {
    constructor(message, issues) {
        super(message);
        this.issues = issues;
        Object.setPrototypeOf(this, SchemaTypeError.prototype);
    }
}
const assert = (schema, value) => {
    const issues = _validate(schema, value, []);
    if (issues.length !== 0) {
        throw new SchemaTypeError('Invalid type, did not pass schema validation', issues);
    }
};

const is = (schema, value) => _validate(schema, value, []).length === 0;

exports.any = any;
exports.array = array;
exports.assert = assert;
exports.bigint = bigint;
exports.boolean = boolean;
exports.code = code$j;
exports.function = functionType;
exports.intersection = intersection;
exports.is = is;
exports.literal = literal;
exports.literalUnion = literalUnion;
exports.null = nullType;
exports.number = number;
exports.object = object;
exports.optional = optional;
exports.readonly = readonly;
exports.record = record;
exports.string = string;
exports.tuple = tuple;
exports.undefined = undefined$1;
exports.union = union;
exports.unknown = unknown;
exports.validate = validate$k;
exports.void = voidType;
//# sourceMappingURL=index.js.map
